#include "decon_sac.h"

/* decon_sac: takes a SAC file as input and deconvolves the instrument
              response (which is contained in the corresponding RESP file)
              from the data in the SAC file to obtain the displacement,
              velocity, or acceleration recorded by the instrument.
   
      Usage: decon_sac -f infilename [-u (DIS | VEL | ACC)] [-wl waterlevel]
   
              If the user doesn't set a value for the units, then the
              displacement (DIS) will be output.  If a waterlevel is not
              set by the user then a waterlevel of 1.0e-04 will be used
              in the deconvolution.
   
              In the output SAC file, the component name is changed by appending
              the string "_DIS", "_VEL", or "_ACC" to the component name in the
              input SAC header.  The output SAC file is created with a name similar
              in form to the SAC filenames generated by rdseed.  An example input
              and output filename are shown below:
   
      1995.211.05.08.40.1540.IU.ANMO.LHZ.SAC     - the input filename
      1995.211.05.08.40.1540.IU.ANMO.LHZ_ACC.SAC - the corresponding output filename
                                                   for acceleration.
  
*/

main(int argc, char **argv) {
    FILE *fptr;
    char *out_fname, *filename = (char *)NULL, *tmp_ptr, *datime;
    char *units, *def_units = "DIS";
    char network[NETLEN], time_string[15];
    float *data_ptr, *trans_ptr, *lcl_ptr;
    float mean_val, val, mod_val, real_prod, complex_prod;
    float incr, delta, maxSS, gamma, waterlevel, default_waterlevel = 1.0e-04;
    double exp2, *freqs;
    unsigned long nreal, ncomplex;
    long npts, mid_idx, i, j;
    int pow2, fldlen, len;
    struct evalresp_args args;
    struct sac sac_hdr, *hdr_ptr;
    struct response *first;
    struct complex *resp_ptr;

    /* get the filename and units (if any) from the command line.  If the filename is
       not found, complain and exit.  If no units are given, use a default value of
       "dis" (displacement) in the evalresp command.  If no waterlevel is specified
       then a default value of 1.0e-04 will be used.  */

    if(argc != 3 && argc != 5 && argc != 7) {
        fprintf(stderr,"Usage: %s -f infilename [-u (DIS | VEL | ACC)] [-wl waterlevel]\n",argv[0]);
        exit(-1);
    }
    units = def_units;
    waterlevel = default_waterlevel;
    for(i = 1; i < argc; i+=2) {
        if(!strcmp(argv[i],"-f"))
            filename = argv[i+1];
        else if(!strcmp(argv[i],"-u")) {
            if(strcmp(argv[i+1],"DIS") && strcmp(argv[i+1],"VEL") && strcmp(argv[i+1],"ACC")) {
                fprintf(stderr,"ERROR (%s): unrecognized units given, check usage\n", argv[0]);
                exit(-1);
            }
            units = argv[i+1];
        }
        else if(!strcmp(argv[i],"-wl"))
            waterlevel = (float)atof(argv[i+1]);
        else {
            fprintf(stderr,"ERROR (%s): unrecognized command line option '%s',", argv[0], argv[i]);
            fprintf(stderr," check usage\n");
            exit(-1);
        }
    }
    if(filename == (char *)NULL) {
        fprintf(stderr,"Usage: %s -f infilename [-u (DIS | VEL | ACC)] [-wl waterlevel]\n",argv[0]);
        exit(-1);
    }

    if((fptr = fopen(filename,"r")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], filename);
        exit(-2);
    }

    /* set the 'hdr_ptr' to point to the address of the 'sac_hdr' */

    hdr_ptr = &sac_hdr;   

    /* read the sac header and the data from the input file */

    npts = read_sac(fptr, READ_DATA, hdr_ptr, &data_ptr);
    mean_val = get_mean(data_ptr, npts);
    delta = sac_hdr.delta;
    fclose(fptr);

    /* place the information from the sac header into a structure that
       will be used to construct the call to evalresp */

    strcpy(args.sta, strtok(sac_hdr.kstnm, " "));
    strncpy(network,strtok(sac_hdr.knetwk, " "),NETLEN);
    strcpy(args.chan, strtok(sac_hdr.kcmpnm, " "));
    if(strcmp(network,"-12345")) {
        strcpy(args.net, network);
        sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
    }
    else {
        tmp_ptr = get_netcode(filename);
        if(!strlen(tmp_ptr)) {
            strcpy(args.net, "*");
            sprintf(args.resp_file, "RESP.*.%s.%s", args.sta, args.chan);
	}
        else {
            strcpy(args.net, tmp_ptr);
            sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
            strcpy(sac_hdr.knetwk, tmp_ptr);
        }
    }
    strcpy(args.units, units);
    strcpy(args.resp_type, "cs");
    strcpy(args.spacing, "lin");
    args.year = sac_hdr.nzyear;
    args.jday = sac_hdr.nzjday;
    args.hour = sac_hdr.nzhour;
    args.min = sac_hdr.nzmin;
    args.sec = (float) sac_hdr.nzsec + ((float) sac_hdr.nzmsec) / 1000.;
    args.max_freq = 1.0/(2.0*delta);

    /* determine the next higher power of two after 'npts' */

    exp2 = log((double)npts)/log(2.0);
    pow2 = ((int)exp2)+1;
    ncomplex = (int)pow(2.0,(double)pow2);
    incr = 1.0/(ncomplex*delta);
    args.min_freq = -args.max_freq + incr;

    /* set the number of frequencies to determine the instrument response
       for to half the next higher power of two and set up the array of frequencies
       to evaluate the response at */

    args.nfreqs = ncomplex;
    if((freqs = (double *)malloc(ncomplex*sizeof(double))) == (double *)NULL) {
        fprintf(stderr,"ERROR (%s): malloc failed for freqs array\n", argv[0]);
        exit(-3);
    }
    mid_idx = ncomplex/2;
    for(i = 0, val = 0.0; i <= mid_idx; i++, val += incr) {
        freqs[i] = val;
    }
    for(val = args.min_freq; i < ncomplex; i++, val += incr) {
        freqs[i] = val;
    }

    /* allocate space for the complex vector needed to hold the data.
       Note:  the size of this array must be a power of 2 */

    nreal = ncomplex * 2;
    if((trans_ptr = (float *)malloc(nreal*sizeof(float))) == (float *)NULL) {
        fprintf(stderr,"ERROR (%s): malloc failed for complex array\n", argv[0]);
        exit(-3);
    }
    for(i = 0; i < npts; i++) {
        *(data_ptr+i) -= mean_val;
    }
    taper(data_ptr, npts, 0.1);
    for(i = 0, j = 0; i < npts; i++, j+=2) {
        *(trans_ptr+j) = *(data_ptr+i);
    }
    for( ; j < nreal; j++) {
        *(trans_ptr+j) = 0.0;
    }

    /* fft the data */

    lcl_ptr = trans_ptr - 1;
    four1(lcl_ptr, ncomplex, -1);

    /* construct and run the evalresp command */

    sprintf(time_string,"%2.2d:%2.2d:%08.5f", args.hour, args.min, args.sec);
    fldlen = strlen(time_string)+9;
    if((datime = (char *)malloc(fldlen)) == (char *)NULL) {
        fprintf(stderr,"ERROR (%s): malloc failed for datime array\n", argv[0]);
        exit(-3);
    }
    sprintf(datime, "%4.4d,%3.3d,%s", args.year, args.jday, time_string);
    first = evresp(args.sta,args.chan,args.net,"*",datime,args.units,(char *)NULL,
                   freqs,args.nfreqs,args.resp_type,(char *)NULL,-1,0,0);

    if(first == (struct response *)NULL) {
        fprintf(stderr,"ERROR (%s): no matching responses found\n", argv[0]);
        exit(-4);
    }
    resp_ptr = first->rvec;

    for(i = 0, j = 0; i < args.nfreqs; i++, j+=2) {
        complex_prod = pow(resp_ptr[i].real,2.0) + pow(resp_ptr[i].imag,2.0);
        maxSS = MAX(maxSS,complex_prod);
    }
    gamma = maxSS * waterlevel;

    /* deconvolve the instrument response (in the vector 'resp_ptr') from the calculated spectra
       (in the vector 'trans_ptr').  Note:  this algorithm uses a fixed waterlevel of 1e-06 */

    for(i = 0, j = 0; i < args.nfreqs; i++, j+=2) {
        real_prod = *(trans_ptr+j) * resp_ptr[i].real + *(trans_ptr+j+1) * resp_ptr[i].imag;
        complex_prod = -*(trans_ptr+j) * resp_ptr[i].imag + *(trans_ptr+j+1) * resp_ptr[i].real;
        mod_val = pow(resp_ptr[i].real,2.0) + pow(resp_ptr[i].imag,2.0);
        *(trans_ptr+j) = real_prod / MAX(mod_val,gamma);
        *(trans_ptr+j+1) = complex_prod / MAX(mod_val,gamma);
    }

    /* inverse fft the complex spectrum */

    four1(lcl_ptr, ncomplex, 1);

    /* change the channel name in the SAC header to let other programs know
       that the instrument response has been deconvolved from this waveform */

    strcat(hdr_ptr->kcmpnm,"_");
    strcat(hdr_ptr->kcmpnm,units);

    /* construct the 'new' output filenames and open the files */
 
    len = 30 + strlen(hdr_ptr->kcmpnm) + strlen(hdr_ptr->kstnm) + strlen(hdr_ptr->knetwk);
    if((out_fname = (char *)malloc(len*sizeof(char))) == (char *)NULL) {
        fprintf(stderr,"ERROR (%s): malloc failed for output filename\n", argv[0]);
        exit(-6);
    }
    sprintf(out_fname,"%4.4d.%3.3d.%2.2d.%2.2d.%2.2d.%4.4d.%s.%s.%s.SAC", hdr_ptr->nzyear,
             hdr_ptr->nzjday, hdr_ptr->nzhour, hdr_ptr->nzmin, hdr_ptr->nzsec,
             (hdr_ptr->nzmsec*10), hdr_ptr->knetwk, hdr_ptr->kstnm, hdr_ptr->kcmpnm);
    if((fptr = fopen(out_fname,"w")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], out_fname);
        exit(-7);
    }

    /* and output the deconvolved waveform (as a SAC file) to stdout */

    for(i = 0, j = 0; i < npts; i++, j+=2) {
        *(data_ptr+i) = *(trans_ptr+j)/((float) ncomplex);
    }
    write_sac(fptr, hdr_ptr, data_ptr); 

    /* and exit with a normal status */

    exit(0);

}
