#include <stdio.h>
#include "sac.h"

/* read_sac:  reads a SAC file and returns the number of points read.

       Arguments:
          FILE *fptr          (input)
          int read_data_flag  (input)
          struct sac *sac_hdr (output)
          float **data_addr   (output)

       Note:  the sac_hdr argument should point to a pre-allocated SAC
              header structure, while the address pointed to by the
              data_addr output argument will be set to the address of
              an array that is allocated at run time by the function.
              In addition, if the read_data_flag is set to zero, only
              the header will be read from the input file. */

int read_sac(FILE *fptr, int read_data_flag, struct sac *sac_hdr, float **data_addr) {
    int npts;
    float *data_ptr;

    if(fread(sac_hdr, sizeof(struct sac), 1, fptr) != 1) {
        fprintf(stderr,"ERROR (read_sac): while reading SAC header from input file\n");
        exit(-2);
    }
    if(!read_data_flag) {
        return(0);
    }
    npts = sac_hdr->npts;
    if((data_ptr = (float *)malloc(npts*sizeof(float))) == (float *)NULL) {
        fprintf(stderr,"ERROR (read_sac): malloc failed for data array with %d elements\n",
                        npts);
        exit(-3);
    }
    if(fread(data_ptr, sizeof(float), npts, fptr) != npts) {
        fprintf(stderr,"ERROR (read_sac): while reading SAC data from input file\n");
        exit(-2);
    }
    *data_addr = data_ptr;
    return(npts);
}

/* write_sac:  writes a SAC file and returns a flag indicating success or failure.

       Arguments:
          FILE *fptr          (input)
          struct sac *sac_hdr (input)
          float *data_ptr   (input)

*/

int write_sac(FILE *fptr, struct sac *sac_hdr, float *data_ptr) {
    int npts;

    if(fwrite(sac_hdr, sizeof(struct sac), 1, fptr) != 1) {
        fprintf(stderr,"ERROR (write_sac): while writing SAC header to output file\n");
        exit(-2);
    }
    npts = sac_hdr->npts;
    if(fwrite(data_ptr, sizeof(float), npts, fptr) != npts) {
        fprintf(stderr,"ERROR (write_sac): while writing SAC data to output file\n");
        exit(-2);
    }
}
