#include "sac_rot.h"

/* global variables that are used to define the set of orientation codes
   that are legal for horizontal components (according to the SEED manual).
   Note:  the values of legal_east and legal_north must be seperated by a
          "|" character (since this is the token that will be used to
          parse these strings into pairs of components) */

char *legal_east = "E|W";
char *legal_north = "N|V";
char *legal_pairs = "(N and E) | (V and W)";
char *east_pattern = "^[EW]$";
char *north_pattern = "^[NV]$";

/* sac_rot: takes pair of 'non-traditionally' oriented horizontal component SAC
            files as input and generates a corresponding pair of SAC files with
            the channels oriented in the 'traditional' north and east directions.
 
    Usage: sac_rot -c1 comp1_file -c2 comp2_file -noc [NV] -eoc [EW]
 
            Here, 'comp1_file' and 'comp2_file' are the names of the input files,
            while the values that follow the '-noc' and '-eoc' flags are the
            "north" and "east" orientation codes that will be used to built the
            component names for the rotated channels.  These values must follow
            the SEED conventions for channel orientation codes.  A subset of the
            possible channel orientation code values are shown above (note, in the
            SEED convention, W is an optional east component, not a west component).
            Only logical pairs of orientation codes are allowed (N and E or V and W).
            If a 'non-logical' pair (like N and W) is chosen, then an error will
            result.  Similarly, an error will result if a 'non-north' ('non-east')
            channel, like 'W' ('V'), is input as the value for the north (east)
            orientation code.
 
            When rotating the waveforms, only waveforms whose sample rates match
            may be rotated.  In addition, the start times of the overlapping data
            segments may not differ by more than 10% of the sample interval and the
            component azimuth values in the SAC headers of the two files must not
            be equal.
 
            If the conditions stated above are satisfied, then this algorithm will
            determine which segments of data overlap as follows:
 
            (1) compare the start times of the two records.  Take the maximum of the
                two start times as the start time for the rotated waveforms.
            (2) if one record starts earlier than the other, increment the pointer to
                that record and decrement the number of points in that record by the
                difference (in samples) between the start of the 'overlapping' data
                segment and the start of the earlier record.
            (3) determine the minimum number of points in the two vectors.  This will
                become the number of points in the rotated waveforms.
 
            Once the 'overlapping' data segments have been rotated to form the "north"
            and "east" data channels, this routine will change the SAC headers for the
            rotated waveforms so that they refer to the new record start time and number
            of points.  In addition, the component azimuth and component name will be
            changed to refer to the new component's name and azimuth.  Following the
            SEED conventions for channel naming, this algorithm simply replaces the
            third letter of the existing channel names (taken from the SAC headers that
            were read from 'comp1_file' and 'comp2_file') with the letter that the user
            entered for the "north" and "east" orientation codes on the command line.
            The azimuth in the "north" channel's SAC header is set to 0 degrees, while
            the azimuth in the "east" channel's SAC header is set to 90 degrees.  This
            algorithm does take into account possible polarity reversals due to the effects
            of a negative overall gain.

*/

main(int argc, char **argv) {
    FILE *out_stream, *fptr1, *fptr2, *out_fptr;
    char *file1 = (char *)NULL, *file2 = (char *)NULL, *north_file, *east_file;
    char *north_chan, *east_chan, *test, *tmp_ptr, *tmp_noc, *tmp_eoc;
    char tmp_str1[64], tmp_str2[64], *tmp_ptr1 = tmp_str1, *tmp_ptr2 = tmp_str2;
    char *north_fname, *east_fname, *def_noc = "N", *def_eoc = "E";
    char tmp_line[MAXLINELEN], network[NETLEN];
    char cmd_string[MAXLINELEN], time_string[15];
    float *north_data, *east_data, *data_ptr1, *data_ptr2;
    float *new_east, *new_north, delta1, delta2, delta;
    float chan1_A0 = 1.0, chan1_sensit = 1.0, radial;
    float chan2_A0 = 1.0, chan2_sensit = 1.0, transverse;
    float north_A0, north_sensit, north_az, tmp_val;
    float east_A0, east_sensit, east_az, mean_val;
    double north_start, east_start;
    double phi1, phi2, alpha, n_scale, e_scale;
    double cosphi1, cosphi2, sinphi1, sinphi2;
    double tolerance = 0.1, baz;
    int north_idx;
    long east_npts, north_npts, npts1, npts2, diff_npts, i, j;
    long npts, e_start, n_start, len;
    struct sac sac_hdr1, sac_hdr2;
    struct sac *hdr_ptr1, *hdr_ptr2, *north_hdr, *east_hdr;
    struct evalresp_args args;
    BTIME time;

    pi = acos(-1.0);
    rad_per_deg = pi / 180.0;

    /* get the filenames from the command line.  If an input filename is not
       found, complain and exit */

    if(argc != 5 && argc != 9) {
        fprintf(stderr,"Usage: %s -c1 comp1_file -c2 comp2_file [-noc (%s)] [-eoc (%s)]\n",argv[0],
                 legal_north, legal_east);
        exit(-1);
    }
    east_chan = def_eoc;
    north_chan = def_noc;
    for(i = 1; i < argc; i+=2) {
        if(!strcmp(argv[i],"-c1")) {
            file1 = argv[i+1];
        }
        else if(!strcmp(argv[i],"-c2")) {
            file2 = argv[i+1];
        }
        else if(!strcmp(argv[i],"-eoc")) {
            east_chan = argv[i+1];
            if(!match_regexp(east_chan,east_pattern)) {
                fprintf(stderr,"illegal east channel orientation code given (%s),", east_chan);
                fprintf(stderr," check usage for legal values\n");
                exit(-1);
            }
        }
        else if(!strcmp(argv[i],"-noc")) {
            north_chan = argv[i+1];
            if(!match_regexp(north_chan,north_pattern)) {
                fprintf(stderr,"illegal north channel orientation code given (%s),", north_chan);
                fprintf(stderr," check usage for legal values\n");
                exit(-1);
            }
        }
        else {
            fprintf(stderr,"ERROR (%s): unrecognized option '%s', check usage\n",
                           argv[0], argv[i]);
            exit(-1);
        }
    }

    if(file1 == (char *)NULL || file2 == (char *)NULL || !strcmp(file1,file2)) {
        fprintf(stderr,"ERROR (%s): two input filenames required, check usage\n",
                       argv[0]);
        exit(-1);
    }
    strncpy(tmp_ptr1, legal_north, 64);
    strncpy(tmp_ptr2, legal_east, 64);
    tmp_noc = strtok(tmp_ptr1,"|");
    tmp_eoc = strtok(tmp_ptr2,"|");
    while(tmp_noc != (char *)NULL) {
        if(!strcmp(north_chan,tmp_noc) && strcmp(east_chan,tmp_eoc)) {
            fprintf(stderr,"ERROR (%s): inconsistent north and east orientation codes",
                           argv[0]);
            fprintf(stderr," channels\n  should consist of one of the following pairs:\n");
            fprintf(stderr,"          %s\n", legal_pairs);
            exit(-1);
        }
        tmp_ptr1 += strlen(tmp_noc)+1;
        tmp_noc = strtok(tmp_ptr1,"|");
        tmp_ptr2 += strlen(tmp_eoc)+1;
        tmp_eoc = strtok(tmp_ptr2,"|");
    }

    if((fptr1 = fopen(file1,"r")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], file1);
        exit(-2);
    }
    if((fptr2 = fopen(file2,"r")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], file2);
        exit(-2);
    }

    /* set the 'hdr_ptr1' to point to the address of the 'sac_hdr1' and 'hdr_ptr2'
       to point to the address of the 'sac_hdr2' */

    hdr_ptr1 = &sac_hdr1;   
    hdr_ptr2 = &sac_hdr2;   

    /* read the sac headers and the data from the input files */

    npts1 = read_sac(fptr1, READ_DATA, hdr_ptr1, &data_ptr1);
    npts2 = read_sac(fptr2, READ_DATA, hdr_ptr2, &data_ptr2);
    fclose(fptr1);
    fclose(fptr2);

    /* if the sample rates differ, exit with an error */

    delta1 = hdr_ptr1->delta;
    delta2 = hdr_ptr2->delta;
    if(delta1!=delta2) {
        fprintf(stderr,"ERROR (%s): cannot rotate when sample rates differ\n", argv[0]);
        exit(-5);
    }

    /* chose one of the two sample rates as the 'new' sample rate */

    delta = delta1;

    /* demean the traces */

    mean_val = get_mean(data_ptr1, npts1);
    for(i = 0; i < npts1; i++) {
        *(data_ptr1+i) -= mean_val;
    }
    mean_val = get_mean(data_ptr2, npts2);
    for(i = 0; i < npts2; i++) {
        *(data_ptr2+i) -= mean_val;
    }

    /* place the information from the 'north' sac header into a structure that
       will be used to construct the call to evalresp */

    strcpy(args.sta, strtok(hdr_ptr1->kstnm, " "));
    strncpy(network,strtok(hdr_ptr1->knetwk, " "),NETLEN);
    strcpy(args.chan, strtok(hdr_ptr1->kcmpnm, " "));
    if(strcmp(network,"-12345")) {
        strcpy(args.net, network);
        sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
    }
    else {
        tmp_ptr = get_netcode(file1);
        strcpy(args.net, tmp_ptr);
        if(!strlen(tmp_ptr))
            sprintf(args.resp_file, "RESP.*.%s.%s", args.sta, args.chan);
        else {
            sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
            strcpy(hdr_ptr1->knetwk,tmp_ptr);
        }
    }
    strcpy(args.units, "def");
    strcpy(args.resp_type, "cs");
    strcpy(args.spacing, "lin");
    args.year = hdr_ptr1->nzyear;
    args.jday = hdr_ptr1->nzjday;
    args.hour = hdr_ptr1->nzhour;
    args.min = hdr_ptr1->nzmin;
    args.sec = (float) hdr_ptr1->nzsec + ((float) hdr_ptr1->nzmsec) / 1000.;
    args.max_freq = 0.5 * 1.0/(2.0*hdr_ptr1->delta);
    args.min_freq = 0.5 * 1.0/(2.0*hdr_ptr1->delta);
    args.nfreqs = 1;

    /* construct and run the evalresp command */

    sprintf(time_string,"%2.2d:%2.2d:%08.5f", args.hour, args.min, args.sec);
    if(strlen(args.net)) {
      sprintf(cmd_string,"%s %s %s %d %d %f %f %d -t %s -n %s -u %s -r %s -s %s %s %s %s",
                 "evalresp", args.sta, args.chan, args.year, args.jday, args.min_freq,
                 args.max_freq, args.nfreqs, time_string, args.net, args.units,
                 args.resp_type, args.spacing, "-v -stdio < ", args.resp_file,
                 "2>&1 | cat");
    }
    else {
      sprintf(cmd_string,"cat %s | %s %s %s %d %d %f %f %d -t %s -u %s -r %s %s %s %s",
                 args.resp_file, "evalresp", args.sta, args.chan, args.year, args.jday,
                 args.min_freq, args.max_freq, args.nfreqs, time_string, args.units,
                 args.resp_type, args.spacing, "-v -stdio", "2>&1 | cat");
    }
    out_stream = popen(cmd_string,"r");

    /* parse the output from evalresp retrieve the sensitivity and A0 values for the
       'north' channel */

    while((test = fgets(tmp_line,MAXLINELEN,out_stream)) != (char *)NULL) {
        if(get_float(tmp_line, "^   computed sens=([-+0-9.E]*)", &tmp_val))
            chan1_sensit = tmp_val;
        else if(get_float(tmp_line, "^     stage  1.*A0=([-+0-9.E]*)", &tmp_val))
            chan1_A0 = tmp_val;
    }
    pclose(out_stream);

    /* place the information from the 'east' sac header into a structure that
       will be used to construct the call to evalresp */

    strcpy(args.sta, strtok(hdr_ptr2->kstnm, " "));
    strncpy(network,strtok(hdr_ptr2->knetwk, " "),NETLEN);
    strcpy(args.chan, strtok(hdr_ptr2->kcmpnm, " "));
    if(strcmp(network,"-12345")) {
        strcpy(args.net, network);
        sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
    }
    else {
        tmp_ptr = get_netcode(file2);
        strcpy(args.net, tmp_ptr);
        if(!strlen(tmp_ptr))
            sprintf(args.resp_file, "RESP.*.%s.%s", args.sta, args.chan);
        else {
            sprintf(args.resp_file, "RESP.%s.%s.%s", args.net, args.sta, args.chan);
            strcpy(hdr_ptr2->knetwk,tmp_ptr);
        }
    }
    strcpy(args.units, "def");
    strcpy(args.resp_type, "cs");
    strcpy(args.spacing, "lin");
    args.year = hdr_ptr2->nzyear;
    args.jday = hdr_ptr2->nzjday;
    args.hour = hdr_ptr2->nzhour;
    args.min = hdr_ptr2->nzmin;
    args.sec = (float) hdr_ptr2->nzsec + ((float) hdr_ptr2->nzmsec) / 1000.;
    args.max_freq = 0.5 * 1.0/(2.0*hdr_ptr2->delta);
    args.min_freq = 0.5 * 1.0/(2.0*hdr_ptr2->delta);
    args.nfreqs = 1;

    /* construct and run the evalresp command */

    sprintf(time_string,"%2.2d:%2.2d:%08.5f", args.hour, args.min, args.sec);
    if(strlen(args.net)) {
      sprintf(cmd_string,"%s %s %s %d %d %f %f %d -t %s -n %s -u %s -r %s -s %s %s %s %s",
                 "evalresp", args.sta, args.chan, args.year, args.jday, args.min_freq,
                 args.max_freq, args.nfreqs, time_string, args.net, args.units,
                 args.resp_type, args.spacing, "-v -stdio < ", args.resp_file,
                 "2>&1 | cat");
    }
    else {
      sprintf(cmd_string,"cat %s | %s %s %s %d %d %f %f %d -t %s -u %s -r %s %s %s %s",
                 args.resp_file, "evalresp", args.sta, args.chan, args.year, args.jday,
                 args.min_freq, args.max_freq, args.nfreqs, time_string, args.units,
                 args.resp_type, args.spacing, "-v -stdio", "2>&1 | cat");
    }
    out_stream = popen(cmd_string,"r");

    /* parse the output from evalresp retrieve the sensitivity and A0 values for the
       'east' channel */

    while((test = fgets(tmp_line,MAXLINELEN,out_stream)) != (char *)NULL) {
        if(get_float(tmp_line, "^   computed sens=([-+0-9.E]*)", &tmp_val))
            chan2_sensit = tmp_val;
        else if(get_float(tmp_line, "^     stage  1.*A0=([-+0-9.E]*)", &tmp_val))
            chan2_A0 = tmp_val;
    }
    pclose(out_stream);

    /* check for negative sensitivities.  If found, rotate component by 180 degrees
       and reset sensitivity to a positive value */

    if(chan1_sensit < 0) {
        sac_hdr1.cmpaz = (int)fmod(((double)sac_hdr1.cmpaz+180.0),360.0);
        chan1_sensit = -chan1_sensit;
    }
    if(chan2_sensit < 0) {
        sac_hdr2.cmpaz = (int)fmod(((double)sac_hdr2.cmpaz+180.0),360.0);
        chan2_sensit = -chan2_sensit;
    }

    /* check to see which is 'north' and which is 'east' */

    north_idx = find_north(sac_hdr1.cmpaz, sac_hdr2.cmpaz);

    if(north_idx == 2) {                /* 'c2' is 'north' */
        north_data = data_ptr2;
        north_hdr = hdr_ptr2;
        north_npts = npts2;
        north_A0 = chan2_A0;
        north_sensit = chan2_sensit;
        north_az = sac_hdr2.cmpaz;
        north_file = file2;
        east_data = data_ptr1;
        east_hdr = hdr_ptr1;
        east_npts = npts1;
        east_A0 = chan1_A0;
        east_sensit = chan1_sensit;
        east_az = sac_hdr1.cmpaz;
        east_file = file1;
    }
    else if(north_idx == 1) {
        north_data = data_ptr1;
        north_hdr = hdr_ptr1;
        north_npts = npts1;
        north_A0 = chan1_A0;
        north_sensit = chan1_sensit;
        north_az = sac_hdr1.cmpaz;
        north_file = file1;
        east_data = data_ptr2;
        east_hdr = hdr_ptr2;
        east_npts = npts2;
        east_A0 = chan2_A0;
        east_sensit = chan2_sensit;
        east_az = sac_hdr2.cmpaz;
        east_file = file2;
    }
    else {
        fprintf(stderr,"ERROR (%s): cannot rotate parallel components\n", argv[0]);
        exit(-5);
    }

    /* check the start times of the 'north' and 'east' vectors.
       subtract the difference in the number of points from the
       record with the minimum start time and reset the pointer.
       then adjust the number of points to be the minimum of the
       two vector lengths */

    time.year = north_hdr->nzyear;
    time.day = north_hdr->nzjday;
    time.hours = north_hdr->nzhour;
    time.minutes = north_hdr->nzmin;
    time.seconds = north_hdr->nzsec;
    time.frac_secs = north_hdr->nzmsec * 10;
    north_start = btime2epoch(&time);

    time.year = east_hdr->nzyear;
    time.day = east_hdr->nzjday;
    time.hours = east_hdr->nzhour;
    time.minutes = east_hdr->nzmin;
    time.seconds = east_hdr->nzsec;
    time.frac_secs = east_hdr->nzmsec * 10;
    east_start = btime2epoch(&time);

    new_north = north_data;
    new_east = east_data;

    if(north_start < east_start) {
        diff_npts = (int)(0.5 + (east_start-north_start)/(double)delta);
        if(diff_npts >= north_npts) {
            fprintf(stderr,"ERROR (%s) cannot rotate, non-overlapping data segments\n", argv[0]);
            fprintf(stderr,"    east channel starts %f seconds after north channel ends\n",
                    ((float)(diff_npts-north_npts))*delta);
            exit(-5);
        }
        north_npts -= diff_npts;
        north_start += (diff_npts * delta);
        new_north += diff_npts;
        epoch2btime(&time, east_start);
    }
    else if (east_start < north_start) {
        diff_npts = (int)(0.5 + (north_start-east_start)/(double)delta);
        if(diff_npts >= east_npts) {
            fprintf(stderr,"ERROR (%s) cannot rotate, non-overlapping data segments\n", argv[0]);
            fprintf(stderr,"    north channel starts %f seconds after east channel ends\n",
                    ((float)(diff_npts-east_npts))*delta);
            exit(-5);
        }
        east_npts -= diff_npts;
        east_start += (diff_npts * delta);
        new_east += diff_npts;
        epoch2btime(&time, north_start);
    }

    if(ABS(north_start-east_start) > (tolerance*delta)) {
        fprintf(stderr,"ERROR (%s): cannot rotate when start times differ by more than %f secs\n",
                   argv[0], (tolerance*delta));
        exit(-5);
    }

    npts = MIN(east_npts,north_npts);

    /* rotate the two waveforms into 'standard' orientations.  To do this, set the
       'backazimuth' to 180 degrees (pi radians) and use a 'standard' rotation from
       'east' and 'north' to 'radial' and 'transverse'.  Then the 'radial' component
       is positive away from the source (i.e. to the north), while the 'transverse'
       component is positive in a 'clockwise' direction form the source (i.e. to the
       east).  Then the 'radial' component corresponds to the desired 'north' channel
       while the 'transverse' component corresponds to the desired 'east' channel */

    baz = pi;
    n_scale = (north_A0 * north_sensit) / (east_A0 * east_sensit);
    e_scale = (east_A0 * east_sensit) / (north_A0 * north_sensit);

    /* determine the phi1 and alpha values (in radians) */

    phi1 = baz - north_az * rad_per_deg;
    alpha = (east_az - north_az - 90.0) * rad_per_deg;

    cosphi1 = cos(phi1);
    sinphi1 = sin(phi1);

    if(alpha == 0.0) {    /* orthogonal components so use 'standard' rotation algorithm */
        for(i = 0; i < npts; i++) {
            radial = -1.0 * (new_north[i]*cosphi1 + new_east[i]*sinphi1*n_scale);
            transverse = -1.0 * (new_east[i]*cosphi1 - new_north[i]*sinphi1*e_scale);
            new_north[i] = radial;                  /* radial is north */
            new_east[i] = transverse;               /* transverse is east */
        }
    }
    else {                /* non-orthogonal so use a 'modified' rotation algorithm */
        phi2 = pi/2.0-phi1+alpha;
        cosphi2 = cos(phi2);
        sinphi2 = sin(phi2);
        for(i = 0; i < npts; i++) {
            radial = -1.0 * (new_north[i]*cosphi1 + new_east[i]*cosphi2*n_scale);
            transverse = -1.0 * (new_east[i]*sinphi2 - new_north[i]*sinphi1*e_scale);
            new_north[i] = radial;                  /* radial is north */
            new_east[i] = transverse;               /* transverse is east */
        }
    }

    /* set up the 'new' SAC headers */

    east_hdr->nzyear = time.year;
    east_hdr->nzjday = time.day;
    east_hdr->nzhour = time.hours;
    east_hdr->nzmin = time.minutes;
    east_hdr->nzsec = time.seconds;
    east_hdr->nzmsec = time.frac_secs / 10;
    east_hdr->npts = npts;
    east_hdr->delta = delta;
    east_hdr->e = (float)(npts-1) * delta;
    east_hdr->cmpaz = 90;
    *(east_hdr->kcmpnm+2) = *east_chan;

    north_hdr->nzyear = time.year;
    north_hdr->nzjday = time.day;
    north_hdr->nzhour = time.hours;
    north_hdr->nzmin = time.minutes;
    north_hdr->nzsec = time.seconds;
    north_hdr->nzmsec = time.frac_secs / 10;
    north_hdr->npts = npts;
    north_hdr->delta = delta;
    north_hdr->e = (float)(npts-1) * delta;
    north_hdr->cmpaz = 0;
    *(north_hdr->kcmpnm+2) = *north_chan;

    /* construct the 'new' output filenames and open the files */

    if(strcmp(north_hdr->knetwk,"-12345")) {
        len = 30 + strlen(north_hdr->kcmpnm) + strlen(north_hdr->kstnm) + strlen(north_hdr->knetwk);
        if((north_fname = (char *)malloc(len*sizeof(char))) == (char *)NULL) {
            fprintf(stderr,"ERROR (%s): malloc failed for north filename\n", argv[0]);
            exit(-6);
        }
        sprintf(north_fname,"%4.4d.%3.3d.%2.2d.%2.2d.%2.2d.%4.4d.%s.%s.%s.SAC", north_hdr->nzyear,
             north_hdr->nzjday, north_hdr->nzhour, north_hdr->nzmin, north_hdr->nzsec,
             (north_hdr->nzmsec*10), north_hdr->knetwk, north_hdr->kstnm, north_hdr->kcmpnm);
    }
    else {
        len = 30 + strlen(north_hdr->kcmpnm) + strlen(north_hdr->kstnm);
        if((north_fname = (char *)malloc(len*sizeof(char))) == (char *)NULL) {
            fprintf(stderr,"ERROR (%s): malloc failed for north filename\n", argv[0]);
            exit(-6);
        }
        sprintf(north_fname,"%4.4d.%3.3d.%2.2d.%2.2d.%2.2d.%4.4d..%s.%s.SAC", north_hdr->nzyear,
             north_hdr->nzjday, north_hdr->nzhour, north_hdr->nzmin, north_hdr->nzsec,
             (north_hdr->nzmsec*10), north_hdr->kstnm, north_hdr->kcmpnm);
    }

    if(strcmp(east_hdr->knetwk,"-12345")) {
        len = 30 + strlen(east_hdr->kcmpnm) + strlen(east_hdr->kstnm) + strlen(east_hdr->knetwk);
        if((east_fname = (char *)malloc(len*sizeof(char))) == (char *)NULL) {
            fprintf(stderr,"ERROR (%s): malloc failed for east filename\n", argv[0]);
            exit(-6);
        }
        sprintf(east_fname,"%4.4d.%3.3d.%2.2d.%2.2d.%2.2d.%4.4d.%s.%s.%s.SAC", east_hdr->nzyear,
             east_hdr->nzjday, east_hdr->nzhour, east_hdr->nzmin, east_hdr->nzsec,
             (east_hdr->nzmsec*10), east_hdr->knetwk, east_hdr->kstnm, east_hdr->kcmpnm);
    }
    else {
        len = 30 + strlen(east_hdr->kcmpnm) + strlen(east_hdr->kstnm);
        if((east_fname = (char *)malloc(len*sizeof(char))) == (char *)NULL) {
            fprintf(stderr,"ERROR (%s): malloc failed for east filename\n", argv[0]);
            exit(-6);
        }
        sprintf(east_fname,"%4.4d.%3.3d.%2.2d.%2.2d.%2.2d.%4.4d..%s.%s.SAC", east_hdr->nzyear,
             east_hdr->nzjday, east_hdr->nzhour, east_hdr->nzmin, east_hdr->nzsec,
             (east_hdr->nzmsec*10), east_hdr->kstnm, east_hdr->kcmpnm);
    }

    /* output the rotated waveforms (as SAC files) */

    if((out_fptr = fopen(north_fname,"w")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], north_fname);
        exit(-7);
    }
    write_sac(out_fptr, north_hdr, new_north);
    fclose(out_fptr);

    if((out_fptr = fopen(east_fname,"w")) == (FILE *)NULL) {
        fprintf(stderr,"ERROR (%s): failed to open file '%s'\n", argv[0], east_fname);
        exit(-7);
    }
    write_sac(out_fptr, east_hdr, new_east);
    fclose(out_fptr);


    /* and exit with a normal status */

    exit(0);

}
